@@ -54,13 +54,13 @@ void ForestClassification::initInternal() {
   }
 
   // Set minimal node size
-  if (min_node_size == 0) {
-    min_node_size = DEFAULT_MIN_NODE_SIZE_CLASSIFICATION;
+  if (min_node_size.size() == 1 && min_node_size[0] == 0) {
+    min_node_size[0] = DEFAULT_MIN_NODE_SIZE_CLASSIFICATION;
   }
 
   // Set minimal bucket size
-  if (min_bucket == 0) {
-    min_bucket = DEFAULT_MIN_BUCKET;
+  if (min_bucket.size() == 1 && min_bucket[0] == 0) {
+    min_bucket[0] = DEFAULT_MIN_BUCKET;
   }
 
   // Create class_values and response_classIDs
