@@ -87,6 +87,17 @@ private:
   void findBestSplitValueExtraTreesUnordered(size_t nodeID, size_t varID, size_t num_classes,
       const std::vector<size_t>& class_counts, size_t num_samples_node, double& best_value, size_t& best_varID,
       double& best_decrease);
+  
+  void findBestSplitValueNanSmallQ(size_t nodeID, size_t varID, size_t num_classes,
+                                const std::vector<size_t>& class_counts, size_t num_samples_node, double& best_value, size_t& best_varID,
+                                double& best_decrease);
+  void findBestSplitValueNanSmallQ(size_t nodeID, size_t varID, size_t num_classes,
+                                const std::vector<size_t>& class_counts, size_t num_samples_node, double& best_value, size_t& best_varID,
+                                double& best_decrease, const std::vector<double>& possible_split_values, std::vector<size_t>& counter_per_class,
+                                std::vector<size_t>& counter);
+  void findBestSplitValueNanLargeQ(size_t nodeID, size_t varID, size_t num_classes,
+                                const std::vector<size_t>& class_counts, size_t num_samples_node, double& best_value, size_t& best_varID,
+                                double& best_decrease);
 
   void addImpurityImportance(size_t nodeID, size_t varID, double decrease);
 
